/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.internal;

import java.net.InetSocketAddress;
import java.util.concurrent.ConcurrentHashMap;

import com.aquenos.epics.jackie.client.ChannelAccessClientConfiguration;
import com.aquenos.epics.jackie.client.ClientThreadingStrategy;
import com.aquenos.epics.jackie.client.DefaultChannelAccessClient;
import com.aquenos.epics.jackie.client.beacon.BeaconDetector;
import com.aquenos.epics.jackie.client.internal.ChannelAccessPVServerConnection.State;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;

/**
 * <p>
 * Pool of {@link Server Servers} used by a {@link DefaultChannelAccessClient}.
 * This class is only intended for use by the classes inside this package and by
 * the {@link DefaultChannelAccessClient}.
 * </p>
 * 
 * <p>
 * This class implements a lock-free pool that manages servers. It is used to
 * ensure that there is only one active {@link Server} instance for server
 * socket address. This helps with saving resources as different channels
 * residing on the same server will share a single connection.
 * </p>
 * 
 * <p>
 * Internally, this pool keeps its servers in a {@link ConcurrentHashMap}.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class ServerPool {

    private BeaconDetector beaconDetector;
    private ChannelAccessClientConfiguration configuration;
    private ConcurrentHashMap<InetSocketAddress, Server> servers = new ConcurrentHashMap<InetSocketAddress, Server>();
    private ClientThreadingStrategy threadingStrategy;

    /**
     * Creates a server pool.
     * 
     * @param beaconDetector
     *            beacon detector used by all servers created by this pool. The
     *            beacon detector is used to detect beacon anomalies which is
     *            needed by the servers' connection watch-dogs.
     * @param configuration
     *            configuration to be used by all servers created by this pool.
     * @param threadingStrategy
     *            threading strategy used by the client. The server pool needs
     *            the threading strategy because it has to register newly
     *            created servers with it. The threading strategy then decides
     *            whether it handles the connection in an existing thread or
     *            creates a new thread for the server.
     */
    public ServerPool(BeaconDetector beaconDetector,
            ChannelAccessClientConfiguration configuration,
            ClientThreadingStrategy threadingStrategy) {
        assert (beaconDetector != null);
        assert (configuration != null);
        assert (threadingStrategy != null);
        this.beaconDetector = beaconDetector;
        this.configuration = configuration;
        this.threadingStrategy = threadingStrategy;
    }

    /**
     * Returns a server instance for the specified <code>serverAddress</code>.
     * If an instance for this address already exists within the pool, that
     * instance is returned. Otherwise, a new instance is created. This is
     * handled in a way that guarantees that no active two instances are created
     * in parallel.
     * 
     * @param serverAddress
     *            socket address to be used for the server. This is the address
     *            to which the socket channel that is going to be created will
     *            connect.
     * @param initialVersion
     *            initial Channel Access version to use when creating a new
     *            server. This version is typically determined from the response
     *            to the search requests which led to this server connection to
     *            be created. While protocol upgrades are supported, protocol
     *            downgrades are not. This means that the connection will fail
     *            if the server responds with an older Channel Access version.
     *            If there is already an active server for the specified
     *            <code>serverAddress</code>, this parameter is not used.
     * @return server instance for the specified socket address.
     */
    public Server findOrCreateServer(InetSocketAddress serverAddress,
            ChannelAccessVersion initialVersion) {
        Server server;
        server = servers.get(serverAddress);
        if (server == null) {
            // If no server connection exists yet, we create a new one. When
            // adding the connection, we have to check again whether it already
            // exists because another thread might create a connection
            // concurrently. If the connection still does not exist, we register
            // it with the threading strategy, so that it is processed.
            server = new Server(beaconDetector, configuration, initialVersion,
                    serverAddress, this, threadingStrategy);
            Server existingConnection = servers.putIfAbsent(serverAddress,
                    server);
            if (existingConnection != null) {
                server = existingConnection;
            } else {
                threadingStrategy.addServerConnection(serverAddress, server);
            }
            return server;
        } else {
            // If a connection exists, we have to check whether it has been
            // destroyed. In this case, we remove it and call this method again
            // so that a new connection is created. We do not unregister it from
            // the threading strategy because this is done automatically when
            // the connection is destroyed.
            if (server.getState() == State.CONNECTION_DESTROYED) {
                servers.remove(serverAddress, server);
                return findOrCreateServer(serverAddress, initialVersion);
            } else {
                return server;
            }
        }
    }

    /**
     * Notifies the server pool that a server has been destroyed. This method
     * may only be called by the {@link Server} instance that has been destroyed
     * and is notifying this server pool.
     * 
     * @param serverAddress
     *            socket address of the server.
     * @param server
     *            server instance that has been destroyed.
     */
    void serverDestroyed(InetSocketAddress serverAddress, Server server) {
        // We have to make sure that we only remove the server connection that
        // has been destroyed because a new connection might have been created
        // already.
        servers.remove(serverAddress, server);
        // We do not remove the server from the threading strategy. The server
        // connection will take care of this when it is destroyed.
    }

    /**
     * Destroys this server pool. This destroys all the servers that are
     * currently registered with the server pool.
     */
    public void destroy() {
        // The threading strategy should be destroyed before the server pool is
        // destroyed. This should make sure that no new servers can be created
        // (addServerConnection(...) will throw).
        // Therefore, it is enough if we explicitly destroy all existing
        // servers, just to make sure that the corresponding channels are
        // closed.
        for (Server server : servers.values()) {
            // The server's destroy(...) method should not throw. We set the
            // disconnectImmediately flag to true because the server pool is
            // only destroyed when the whole client is destroyed and in this
            // case the communication threads have already been stopped anyway.
            // This is okay, because a server has to be able to handle the
            // sudden loss of a connection (without the channels being
            // disconnected explicitly) because the same thing will happen if
            // the connection between client and server is interrupted for some
            // reason.
            server.destroy(true);
        }
    }

}
