/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.exception;

import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;

/**
 * <p>
 * {@link BufferUnderflowException} with information about how much data is
 * needed. In contrast to a regular {@link BufferUnderflowException}, this
 * exception indicates how much data is needed in order to perform the operation
 * that just failed.
 * </p>
 * 
 * <p>
 * The number returned is a minimum estimate. This means that the operation
 * might still fail with a BufferUnderflowException if the indicated number of
 * bytes is made available. However, the operation is guaranteed to fail if not
 * at least the indicated number of bytes is made available. Therefore, the
 * calling code does not have to try the operation again until it can supply the
 * required data.
 * </p>
 * 
 * <p>
 * In contrast to most exceptions in this package, this exception does not
 * inherit from {@link JackieRuntimeException}. This is so that it can inherit
 * from {@link BufferUnderflowException} and thus is fully compatible with code
 * that expects a regular {@link BufferUnderflowException}.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class BufferUnderflowWithSizeException extends BufferUnderflowException {

    private static final long serialVersionUID = -4753496009807222169L;

    private int bytesNeeded;

    /**
     * Creates a buffer underflow exception with size information. The specified
     * number of bytes needed is the total number of bytes needed (sum of the
     * number of bytes that already are available and the number of bytes that
     * need to be available in addition to those bytes). In other words, a
     * {@link ByteBuffer}'s {@link ByteBuffer#remaining()} method must at least
     * return the indicated number of bytes for the operation to be able to
     * succeed.
     * 
     * @param bytesNeeded
     *            number of bytes that are at least needed for the operation to
     *            be successful.
     */
    public BufferUnderflowWithSizeException(int bytesNeeded) {
        this.bytesNeeded = bytesNeeded;
    }

    /**
     * <p>
     * Returns the minimum number of bytes that are needed for the operation to
     * succeed. This means that the operation will certainly fail if tried again
     * with less than the indicated number of bytes. However, it might also fail
     * if the indicated number of bytes is available, so the returned number is
     * just a lower estimate.
     * </p>
     * 
     * <p>
     * The number returned is the total number of bytes needed. For example, if
     * this method returns 40 and 24 bytes are already available, 16 bytes need
     * to be supplied in addition to the data that is already available.
     * </p>
     * 
     * @return lower estimate of the number of bytes that are needed for the
     *         operation to succeed.
     */
    public int getBytesNeeded() {
        return bytesNeeded;
    }

}
