/*
 * Copyright 2014 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.io;

import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;

/**
 * <p>
 * Source for reading bytes (and other data types). This is an interface that
 * can be implemented by any class that can provide a stream of data, regardless
 * of where the data comes from. It allows for reading a block of heterogeneous
 * data in a single "transaction", meaning that either the complete block of
 * data is read, or none of it is read and the data stays available for later
 * read operations.
 * </p>
 * <p>
 * If data is read outside of an explicit transaction, each get operation is
 * considered as an individual implicit transaction. If the get operation fails,
 * no data will have been read (and the position within the source will not have
 * changed). If it succeeds, all data will have been read.
 * </p>
 * <p>
 * This interface can be used in a similar way as a {@link ByteBuffer}, however,
 * it is more flexible, because it does not have to be backed by a single
 * object.
 * </p>
 * <p>
 * In general, objects implementing this interface should be considered not
 * thread-safe and if multiple threads access them, the access must be
 * synchronized.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public interface ByteSource {

    /**
     * Interface for use with the
     * {@link ByteSource#atomicGet(AtomicGetOperation)} method. This interface
     * should be implemented by commands that want to read a contiguous block of
     * data in a consistent way.
     * 
     * @param <T>
     *            type of the operations return value. This value is returned to
     *            the code that called
     *            {@link ByteSource#atomicGet(AtomicGetOperation)}.
     * 
     * @author Sebastian Marsching
     */
    static interface AtomicGetOperation<T> {
        /**
         * Method for reading data. If this method throws an exception, all data
         * read from the byte source within this method is not discarded and the
         * source's position is reset to the position before the read operation
         * started. This means that a subsequent read operation will read the
         * data again. If this method returns successfully, all data read by
         * this method is removed from the byte source. The
         * {@link ByteSource#atomicGet(AtomicGetOperation)} method may be called
         * from this method. This will start a nested transaction: If the nested
         * transaction fails (and the exception thrown by the nested transaction
         * is caught within this method), data read within the nested
         * transaction is not discarded. If the nested transaction succeeds but
         * the surrounding transaction fails, the data read by the nested
         * transaction will not be discarded either. Only if both the nested
         * transaction and the surrounding transaction succeed, the data read in
         * the nested transaction is removed from the byte source.
         * 
         * @return result of the get operation (may be empty). This result may
         *         be void. It is subsequently returned by
         *         {@link ByteSource#atomicGet(AtomicGetOperation)}.
         */
        T get();
    }

    /**
     * Starts an atomic get operation. Either all data read within this
     * transaction is discarded (removed from this byte source), or (if the
     * {@link AtomicGetOperation#get()} method of
     * <code>atomicGetOperation</code> throws an exception) all data read within
     * the transaction is kept and this byte source's position is reset to the
     * position before the transaction started. This method will throw any
     * exception thrown by {@link AtomicGetOperation#get()}. Atomic get
     * operations may be nested (see {@link AtomicGetOperation#get()} for
     * details). Despite its name, this method should not be considered to be
     * thread-safe.
     * 
     * @param atomicGetOperation
     *            get operation that shall be performed atomically.
     * @return return value of {@link AtomicGetOperation#get()}.
     * @param <T>
     *            type of the return value of the atomic operation.
     */
    <T> T atomicGet(AtomicGetOperation<T> atomicGetOperation);

    /**
     * Returns the number of bytes remaining in this source. This is the number
     * of bytes that can be read without causing an error. Depending on the
     * implementation of the byte source, there might actually be more data
     * available than indicated by this method. Therefore, generic code that
     * reads from a byte source should not use the return value of this method
     * as an indicator when to stop reading but should rather read until a
     * {@link BufferUnderflowException} is thrown.
     * 
     * @return minimum value of remaining bytes in this byte source (more bytes
     *         than indicated might be available).
     */
    int remaining();

    /**
     * Skips the specified number of bytes. This means that the current position
     * is advanced by the specified number of bytes and the bytes between the
     * old position and the new position are discarded. Subsequent read
     * operations will be relative to the new position.
     * 
     * @param length
     *            number of bytes to be skipped.
     * @return a reference to this byte source.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>length</code> is negative. The current position is
     *             not changed.
     */
    ByteSource skip(int length);

    /**
     * Reads a single byte and advances the current position.
     * 
     * @return single byte read from the current position.
     * @throws BufferUnderflowException
     *             if there are no remaining bytes. The current position is not
     *             changed.
     */
    byte getByte();

    /**
     * Reads the specified number of bytes and advances the current position.
     * This is effectively the same as
     * <code>this.getByteArray(new byte[length], 0, length)</code>.
     * 
     * @param length
     *            number of bytes to be read.
     * @return array with the specified length and containing bytes read from
     *         the source starting at the current position.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>length</code> is negative. The current position is
     *             not changed.
     */
    byte[] getByteArray(int length);

    /**
     * Reads the specified number of bytes into the supplied array and advances
     * the current position.
     * 
     * @param destination
     *            array into which the read data is copied. This array's length
     *            must be at least <code>length + destinationOffset</code>.
     * @param offset
     *            offset into the destination array. The destination array is
     *            filled starting at this position.
     * @param length
     *            number of bytes to be read.
     * @return a reference to this byte source.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IndexOutOfBoundsException
     *             if the destination array is too small. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>offset</code>, <code>length</code> or
     *             <code>destinationOffset</code> is negative. The current
     *             position is not changed.
     */
    ByteSource getByteArray(byte[] destination, int offset, int length);

    /**
     * Reads a single double and advances the current position.
     * 
     * @return single double read from the current position.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     */
    double getDouble();

    /**
     * Reads the specified number of doubles and advances the current position.
     * This is effectively the same as
     * <code>this.getDoubleArray(new double[length], 0, length)</code>.
     * 
     * @param length
     *            number of doubles to be read.
     * @return array with the specified length and containing doubles read
     *         starting at the current position.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>length</code> is negative. The current position is
     *             not changed.
     */
    double[] getDoubleArray(int length);

    /**
     * Reads the specified number of double values into the supplied array and
     * advances the current position.
     * 
     * @param destination
     *            array into which the read values are copied. This array's
     *            length must be at least
     *            <code>length + destinationOffset</code>.
     * @param offset
     *            offset into the destination array. The destination array is
     *            filled starting at this position.
     * @param length
     *            number of values to be read.
     * @return a reference to this byte source.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IndexOutOfBoundsException
     *             if the destination array is too small. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>offset</code>, <code>length</code> or
     *             <code>destinationOffset</code> is negative. The current
     *             position is not changed.
     */
    ByteSource getDoubleArray(double[] destination, int offset, int length);

    /**
     * Reads a single float and advances the current position.
     * 
     * @return single float read from the current position.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     */
    float getFloat();

    /**
     * Reads the specified number of floats and advances the current position.
     * This is effectively the same as
     * <code>this.getFloatArray(new float[length], 0, length)</code>.
     * 
     * @param length
     *            number of floats to be read.
     * @return array with the specified length and containing floats read
     *         starting at the current position.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>length</code> is negative. The current position is
     *             not changed.
     */
    float[] getFloatArray(int length);

    /**
     * Reads the specified number of float values into the supplied array and
     * advances the current position.
     * 
     * @param destination
     *            array into which the read values are copied. This array's
     *            length must be at least
     *            <code>length + destinationOffset</code>.
     * @param offset
     *            offset into the destination array. The destination array is
     *            filled starting at this position.
     * @param length
     *            number of values to be read.
     * @return a reference to this byte source.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IndexOutOfBoundsException
     *             if the destination array is too small. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>offset</code>, <code>length</code> or
     *             <code>destinationOffset</code> is negative. The current
     *             position is not changed.
     */
    ByteSource getFloatArray(float[] destination, int offset, int length);

    /**
     * Reads a single int and advances the current position.
     * 
     * @return single int read from the current position.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     */
    int getInt();

    /**
     * Reads the specified number of ints and advances the current position.
     * This is effectively the same as
     * <code>this.getIntArray(new int[length], 0, length)</code>.
     * 
     * @param length
     *            number of ints to be read.
     * @return array with the specified length and containing ints read starting
     *         at the current position.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>length</code> is negative. The current position is
     *             not changed.
     */
    int[] getIntArray(int length);

    /**
     * Reads the specified number of int values into the supplied array and
     * advances the current position.
     * 
     * @param destination
     *            array into which the read values are copied. This array's
     *            length must be at least
     *            <code>length + destinationOffset</code>.
     * @param offset
     *            offset into the destination array. The destination array is
     *            filled starting at this position.
     * @param length
     *            number of values to be read.
     * @return a reference to this byte source.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IndexOutOfBoundsException
     *             if the destination array is too small. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>offset</code>, <code>length</code> or
     *             <code>destinationOffset</code> is negative. The current
     *             position is not changed.
     */
    ByteSource getIntArray(int[] destination, int offset, int length);

    /**
     * Reads a single long and advances the current position.
     * 
     * @return single long read from the current position.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     */
    long getLong();

    /**
     * Reads the specified number of longs and advances the current position.
     * This is effectively the same as
     * <code>this.getLongArray(new long[length], 0, length)</code>.
     * 
     * @param length
     *            number of longs to be read.
     * @return array with the specified length and containing longs read
     *         starting at the current position.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>length</code> is negative. The current position is
     *             not changed.
     */
    long[] getLongArray(int length);

    /**
     * Reads the specified number of long values into the supplied array and
     * advances the current position.
     * 
     * @param destination
     *            array into which the read values are copied. This array's
     *            length must be at least
     *            <code>length + destinationOffset</code>.
     * @param offset
     *            offset into the destination array. The destination array is
     *            filled starting at this position.
     * @param length
     *            number of values to be read.
     * @return a reference to this byte source.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IndexOutOfBoundsException
     *             if the destination array is too small. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>offset</code>, <code>length</code> or
     *             <code>destinationOffset</code> is negative. The current
     *             position is not changed.
     */
    ByteSource getLongArray(long[] destination, int offset, int length);

    /**
     * Reads a single short and advances the current position.
     * 
     * @return single short read from the current position.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     */
    short getShort();

    /**
     * Reads the specified number of shorts and advances the current position.
     * This is effectively the same as
     * <code>this.getShortArray(new short[length], 0, length)</code>.
     * 
     * @param length
     *            number of shorts to be read.
     * @return array with the specified length and containing shorts read
     *         starting at the current position.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>length</code> is negative. The current position is
     *             not changed.
     */
    short[] getShortArray(int length);

    /**
     * Reads the specified number of short values into the supplied array and
     * advances the current position.
     * 
     * @param destination
     *            array into which the read values are copied. This array's
     *            length must be at least
     *            <code>length + destinationOffset</code>.
     * @param offset
     *            offset into the destination array. The destination array is
     *            filled starting at this position.
     * @param length
     *            number of values to be read.
     * @return a reference to this byte source.
     * @throws BufferUnderflowException
     *             if there are not enough remaining bytes. The current position
     *             is not changed.
     * @throws IndexOutOfBoundsException
     *             if the destination array is too small. The current position
     *             is not changed.
     * @throws IllegalArgumentException
     *             if <code>offset</code>, <code>length</code> or
     *             <code>destinationOffset</code> is negative. The current
     *             position is not changed.
     */
    ByteSource getShortArray(short[] destination, int offset, int length);

}
