/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;

/**
 * Representation of a <code>CA_PROTO_CLEAR_CHANNEL</code> message. This message
 * is sent from a client to a server via TCP in order to indicate that a channel
 * that has been connected previously is not needed any longer and should be
 * disconnected. This means that the resources associated with this channel can
 * be released. Some older servers might send back the same message in order to
 * acknowledge the close operation, but in general no response should be
 * expected.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_CLEAR_CHANNEL
 */
public final class ChannelAccessDisconnectChannelMessage extends
        ChannelAccessMessage {

    private int cid;
    private int sid;

    /**
     * Constructor for creating a <code>CA_PROTO_CLEAN_CHANNEL</code> message.
     * 
     * @param cid
     *            the channel identifier used by the client.
     * @param sid
     *            the channel identifier used by the server.
     */
    public ChannelAccessDisconnectChannelMessage(int cid, int sid) {
        super(ChannelAccessCommand.CA_PROTO_CLEAR_CHANNEL);

        this.cid = cid;
        this.sid = sid;
    }

    private ChannelAccessDisconnectChannelMessage(
            ChannelAccessMessageHeader messageHeader, int cid, int sid) {
        super(ChannelAccessCommand.CA_PROTO_CLEAR_CHANNEL, messageHeader);

        this.cid = cid;
        this.sid = sid;
    }

    /**
     * Returns the channel ID that is used by the client to identify the
     * channel. This integer may be zero or negative.
     * 
     * @return client-side ID for the channel.
     */
    public int getChannelCID() {
        return cid;
    }

    /**
     * Returns the channel ID that is used by the server to identify the
     * channel. This integer may be zero or negative.
     * 
     * @return server-side ID for the channel.
     */
    public int getChannelSID() {
        return sid;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, (short) 0, 0, sid, cid, version,
                maxPayloadSize);
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessDisconnectChannelMessage other = (ChannelAccessDisconnectChannelMessage) obj;
        return new EqualsBuilder().append(this.cid, other.cid)
                .append(this.sid, other.sid).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode()).append(cid)
                .append(sid).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("cid=");
        sb.append(cid);
        sb.append(", sid=");
        sb.append(sid);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete message (including the payload).
     */
    protected static ChannelAccessDisconnectChannelMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_CLEAR_CHANNEL
                .getCommandNumber());
        if (!headerOnly) {
            byteSource.skip(messageHeader.getPayloadSize());
        }
        return new ChannelAccessDisconnectChannelMessage(messageHeader,
                messageHeader.getContextSpecific(), messageHeader.getCID());
    }

}
