/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.Inet4AddressUtil;

/**
 * Representation of a <code>REPEATER_CONFIRM</code> message. This message is
 * sent from the repeater to a client in order to confirm that the client has
 * successfully registered with the repeater.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#REPEATER_CONFIRM
 */
public final class ChannelAccessRepeaterConfirmMessage extends
        ChannelAccessMessage {

    private Inet4Address clientAddress;

    /**
     * Constructor for creating a <code>REPEATER_CONFIRM</code> message. This
     * constructor should be used for messages that are sent from a repeater to
     * a client.
     * 
     * @param clientAddress
     *            the address of the client that has been registered. A client
     *            could use this information to identify which of its
     *            registration attempts have been successful, if it tried to
     *            register with multiple addresses.
     */
    public ChannelAccessRepeaterConfirmMessage(Inet4Address clientAddress) {
        super(ChannelAccessCommand.REPEATER_CONFIRM);

        this.clientAddress = clientAddress;
    }

    private ChannelAccessRepeaterConfirmMessage(
            ChannelAccessMessageHeader messageHeader, Inet4Address clientAddress) {
        super(ChannelAccessCommand.REPEATER_CONFIRM, messageHeader);

        this.clientAddress = clientAddress;
    }

    /**
     * Returns the IP with which the client has been registered. A client could
     * use this information to identify which of its registration attempts have
     * been successful, if it tried to register with multiple addresses.
     * 
     * @return client IP address.
     */
    public Inet4Address getClientAddress() {
        return clientAddress;
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessRepeaterConfirmMessage other = (ChannelAccessRepeaterConfirmMessage) obj;
        return new EqualsBuilder().append(this.clientAddress,
                other.clientAddress).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(clientAddress).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("clientAddress=");
        sb.append(clientAddress != null ? clientAddress.getHostAddress()
                : "null");
        sb.append('}');
        return sb.toString();
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, (short) 0, 0, 0,
                Inet4AddressUtil.addressToInt(clientAddress.getAddress()),
                version, maxPayloadSize);
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete message (including he payload if
     *             <code>headerOnly</code> is <code>false</code>).
     */
    protected static ChannelAccessRepeaterConfirmMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.REPEATER_CONFIRM
                .getCommandNumber());
        try {
            byteSource.skip(messageHeader.getPayloadSize());
        } catch (BufferUnderflowException e) {
            // We ignore a buffer underflow. The repeater protocol uses a single
            // message only per packet anyway, therefore we should always be
            // aligned for the next message, even if the payload size was wrong.
        }
        Inet4Address clientAddress;
        try {
            InetAddress tempAddress = InetAddress.getByAddress(Inet4AddressUtil
                    .addressToBytes(messageHeader.getContextSpecific()));
            if (!(tempAddress instanceof Inet4Address)) {
                throw new RuntimeException(
                        "Did not get IPv4 address where it was expected. Got address of type "
                                + tempAddress.getClass().getName()
                                + " instead.");
            }
            clientAddress = (Inet4Address) tempAddress;
        } catch (UnknownHostException e) {
            throw new RuntimeException(
                    "Cannot get inet address instance for address of four bytes.",
                    e);
        }

        return new ChannelAccessRepeaterConfirmMessage(messageHeader,
                clientAddress);
    }

}
