/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Representation of a <code>CA_PROTO_WRITE_NOTIFY</code> message that is sent
 * from a server to a client. This message is sent in response to a
 * {@link ChannelAccessWriteNotifyClientMessage} in order to notify the client
 * that the write operation has finished.
 * 
 * @see ChannelAccessCommand#CA_PROTO_WRITE_NOTIFY
 * @see ChannelAccessWriteNotifyMessage
 * @see ChannelAccessWriteNotifyClientMessage
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessWriteNotifyServerMessage extends
        ChannelAccessWriteNotifyMessage {

    private short dataType;
    private int count;
    private ChannelAccessStatus status;
    private int operationId;

    /**
     * Creates a <code>CA_PROTO_WRITE_NOTIFY</code> message. This constructor
     * should be used to create a message that is supposed to be sent from a
     * server to a client.
     * 
     * @param dataType
     *            data type sent with the client's write request (never
     *            <code>null</code>).
     * @param count
     *            count sent with the client's request.
     * @param status
     *            status code for the write operation (never <code>null</code>).
     *            If the write was successful, the
     *            {@link ChannelAccessStatus.StatusMessage#ECA_NORMAL
     *            ECA_NORMAL} code should be sent. If an error occurred, the
     *            respective error code should be sent.
     * @param operationId
     *            client-assigned ID identifying the write request. This must be
     *            the ID sent with the write request so that the client can
     *            identify the request to which the reply belongs.
     */
    public ChannelAccessWriteNotifyServerMessage(
            ChannelAccessValueType dataType, int count,
            ChannelAccessStatus status, int operationId) {
        super();

        if (status == null) {
            throw new NullPointerException();
        }

        this.dataType = dataType.toTypeCode();
        this.count = count;
        this.status = status;
        this.operationId = operationId;
    }

    private ChannelAccessWriteNotifyServerMessage(
            ChannelAccessMessageHeader messageHeader, short dataType,
            int count, ChannelAccessStatus status, int operationId) {
        super(messageHeader);

        this.dataType = dataType;
        this.count = count;
        this.status = status;
        this.operationId = operationId;
    }

    /**
     * Returns the data-type that was sent with the client's write request or
     * <code>null</code> if the data-type number could not be converted because
     * it is unknown to this implementation.
     * 
     * @return data-type specified in the write request or <code>null</code> if
     *         the numeric identifier cannot be mapped to a data type known by
     *         this implementation.
     */
    public ChannelAccessValueType getDataType() {
        try {
            return ChannelAccessValueType.forTypeNumber(dataType);
        } catch (IllegalArgumentException e) {
            return null;
        }
    }

    /**
     * Returns the number of value elements that were sent with the client's
     * write request.
     * 
     * @return number of value elements specified in the write request.
     */
    public int getCount() {
        return count;
    }

    /**
     * Returns the status of the write operation. If the write was successful,
     * the status should be {@link ChannelAccessStatus.StatusMessage#ECA_NORMAL}
     * . If the write failed, the status should be the appropriate error code.
     * 
     * @return status code of the write operation. Should be
     *         <code>ECA_NORMAL</code> for a successful write and the error code
     *         for an error.
     */
    public ChannelAccessStatus getStatus() {
        return status;
    }

    /**
     * Returns the client-assigned operation ID. The operation ID is used by the
     * client in order to map replies sent in response to a write request to the
     * respective request. This must always be the ID that was sent by the
     * client. This number might be negative or zero.
     * 
     * @return client-assigned operation identifier.
     */
    public int getOperationId() {
        return operationId;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, dataType, count, status.toStatusCode(),
                operationId, version, maxPayloadSize);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        verifyHeader(0, count, version, maxPayloadSize);
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessWriteNotifyServerMessage other = (ChannelAccessWriteNotifyServerMessage) obj;
        return new EqualsBuilder().append(this.dataType, other.dataType)
                .append(this.count, other.count)
                .append(this.status, other.status)
                .append(this.operationId, other.operationId).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(dataType).append(count).append(status)
                .append(operationId).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("dataType=");
        try {
            sb.append(ChannelAccessValueType.forTypeNumber(dataType));
        } catch (IllegalArgumentException e) {
            sb.append("UNKNOWN(");
            sb.append(dataType);
            sb.append(')');
        }
        sb.append(", count=");
        sb.append(count);
        sb.append(", status=");
        sb.append(status);
        sb.append(", operationId=");
        sb.append(operationId);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     */
    protected static ChannelAccessWriteNotifyServerMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_WRITE_NOTIFY
                .getCommandNumber());
        byteSource.skip(messageHeader.getPayloadSize());
        return new ChannelAccessWriteNotifyServerMessage(messageHeader,
                messageHeader.getDataType(), messageHeader.getCount(),
                ChannelAccessStatus.forStatusCode(messageHeader.getCID()),
                messageHeader.getContextSpecific());
    }

}
