/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import java.nio.charset.Charset;
import java.util.Collection;
import java.util.List;

/**
 * <p>
 * Channel Access values that represents the
 * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} and
 * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} types. This type
 * stores enum states and provides information about the alarm status and enum
 * state labels. In its internal representation, an enum is just an unsigned
 * short integer. However, it is interpreted as a discrete state. In contrast to
 * other types, enum values do not provide additional information for controls
 * type so that the same data structure is used for it.
 * </p>
 * 
 * <p>
 * Due to technical limitations of the Java platform, the enum states are
 * represented by signed integers. The Channel Access protocol, however, defines
 * them as unsigned. Therefore, positive numbers in the Channel Access protocol
 * might appear as negative numbers in this library and vice-versa.
 * </p>
 * 
 * <p>
 * This interface should not be implemented by user-defined classes. Code that
 * wants to create an instance implementing this interface should use the
 * methods provided by {@link ChannelAccessValueFactory} instead.
 * </p>
 * 
 * @author Sebatian Marsching
 */
public interface ChannelAccessGraphicsEnum
        extends ChannelAccessAlarmEnum, ChannelAccessGraphicsOnlyValue<Short>,
        ChannelAccessControlsValue<Short> {

    /**
     * <p>
     * Returns the enum state labels. The list returned contains display labels
     * for up to 16 states. The actual number of labels used is internally
     * stored in the Channel Access data-type. In contrast to the
     * {@link #getRawLabels()} method, this method only returns the labels that
     * have been marked as usable. The list never contains <code>null</code>
     * elements.
     * </p>
     * 
     * <p>
     * Modifications to this list affect this value's labels. The labels raw
     * representation is updated automatically. Therefore, this list may not be
     * modified after this value has been passed to a different component.
     * </p>
     * 
     * <p>
     * If this is a read-only value, the list returned is read-only, too.
     * </p>
     * 
     * @return list with up to 16 labels (never null). The array never contains
     *         <code>null</code> elements.
     * @see #getRawLabels()
     * @see #getCharset()
     */
    List<String> getLabels();

    /**
     * <p>
     * Replaces this value's labels with the elements from the specified
     * collection. This has the same effect as calling <code>clear()</code>
     * followed by <code>addAll(value)</code> on the list returned by
     * {@link #getLabels()}.
     * </p>
     * 
     * <p>
     * The string's raw representation is updated automatically. Strings that
     * are truncated while being converted to their raw representation still
     * maintain their original length in the list of element strings.
     * </p>
     * 
     * @param value
     *            collection of strings that should replace this value's
     *            elements.
     * @throws IllegalArgumentException
     *             if <code>value</code> has more than 16 elements.
     * @throws NullPointerException
     *             if <code>value</code> is <code>null</code> or contains
     *             <code>null</code> elements.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setLabels(Collection<String> value);

    /**
     * <p>
     * Returns the raw representation of the label strings. This an array of 16
     * * 26 (= 416) bytes. Each string in this array has exactly 26 bytes and is
     * null-terminated. This means that the label string with the zero-based
     * index i starts at an array index of i * 26.
     * </p>
     * 
     * <p>
     * The array returned is a copy so modifications to it will not affect this
     * value.
     * </p>
     * 
     * @return raw representation of the label strings (never <code>null</code>
     *         ).
     * @see #getLabels()
     * @see #getCharset()
     */
    byte[] getRawLabels();

    /**
     * <p>
     * Updates the label strings using the specified raw representations. The
     * size of the array passed must be exactly 16 * 26 (= 416) bytes.. Each
     * sequence of 26 bytes is interpreted as one null-terminated string. The
     * strings must be null-terminated, therefore the size of an individual
     * string must not exceed 25 characters (not including the terminating
     * null-byte).
     * </p>
     * 
     * <p>
     * The array is copied, thus subsequent modifications of the array will not
     * affect the labels of this value. The {@link String Java strings} returned
     * by {@link #getLabels()} are updated by converting the specified raw
     * strings using the specified {@link #getCharset() encoding}.
     * </p>
     * 
     * @param rawLabels
     *            new raw representation of the label strings (never
     *            <code>null</code> ). The size of the array must be 16 * 26
     *            bytes and strings must be null terminated.
     * @param numberOfLabels
     *            the number of labels that should be used. Must be a number
     *            between 0 and 16.
     * @throws IllegalArgumentException
     *             if the size of the array is not 416 bytes, if the individual
     *             strings are not null terminated, or if
     *             <code>numberOfLabels</code> is negative or greater than 16.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     * @see #setLabels(Collection)
     * @see #getCharset()
     */
    void setRawLabels(byte[] rawLabels, int numberOfLabels);

    /**
     * Returns the encoding that is used for converting {@linkplain String Java
     * strings} into raw (null-terminated) strings and vice-versa. The encoding
     * is defined when a value instance is created and never changes.
     * 
     * @return encoding used for encoding and decoding strings (never
     *         <code>null</code>).
     */
    Charset getCharset();

    /**
     * <p>
     * Returns a read-only version of this value. If the value is read-only,
     * this value is returned. Otherwise, a wrapped version of this value, that
     * does not allow modifications is returned.
     * </p>
     * 
     * <p>
     * Please note that the read-only value is not a copy of this value, but
     * just a wrapper around it. This means, that modifications of the wrapped
     * value will result in modifications of the read-only value. Therefore it
     * is bad practice to modify a value that has been used as the base for a
     * read-only value, because code that still has a reference to the read-only
     * value might not expect this change and thus fail unexpectedly.
     * </p>
     * 
     * @return read-only version of this value (never <code>null</code>).
     */
    ChannelAccessGraphicsEnum asReadOnlyValue();

    /**
     * <p>
     * Creates and returns a copy of this object. The object returned is
     * completely independent from this object. This means that modifications to
     * this object will not affect the returned object and vice-versa.
     * </p>
     * 
     * <p>
     * The object returned is guaranteed to be completely identical to this
     * object, including its type. This means that
     * <code>x.clone().equals(x)</code> and
     * <code>x.clone().getClass() == x.getClass()</code> are always
     * <code>true</code>.
     * </p>
     * 
     * <p>
     * The returned value is guaranteed to allow write access, even if this
     * value is read-only. This means that cloning a read-only value is a good
     * way to get a writable copy.
     * </p>
     * 
     * @return a clone of this instance.
     */
    ChannelAccessGraphicsEnum clone();

}
