/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

/**
 * <p>
 * Channel Access values that represent a type that stores numbers and provides
 * alarm information, engineering units, and alarm, control, and display limits.
 * </p>
 * 
 * <p>
 * This interface should not be implemented by user-defined classes. Code that
 * wants to create an instance implementing this interface should use the
 * methods provided by {@link ChannelAccessValueFactory} instead.
 * </p>
 * 
 * @param <ElementType>
 *            type of the actual value elements. For primitives, this is the
 *            boxed variant.
 * 
 * @author Sebatian Marsching
 */
public interface ChannelAccessNumericControlsValue<ElementType extends Number>
        extends ChannelAccessControlsValue<ElementType>,
        ChannelAccessNumericGraphicsValue<ElementType> {

    /**
     * <p>
     * Returns the upper control limit. The upper control limit is the upper
     * limit that the server enforces on values. Typically, this means that the
     * process variable cannot be set to a value that is greater than this
     * limit. A value of zero typically indicates that the limit has not been
     * configured and thus should not be used.
     * </p>
     * 
     * <p>
     * Code that wants to get this limit should prefer the type-specific methods
     * that return primitives for better performance. This method is only
     * provided for code that has to deal with values in a generic way (without
     * knowing the exact type of the actual value).
     * </p>
     * 
     * @return upper control limit (never <code>null</code>).
     */
    ElementType getGenericUpperControlLimit();

    /**
     * <p>
     * Returns the lower control limit. The lower control limit is the lower
     * limit that the server enforces on values. Typically, this means that the
     * process variable cannot be set to a value that is less than this limit. A
     * value of zero typically indicates that the limit has not been configured
     * and thus should not be used.
     * </p>
     * 
     * <p>
     * Code that wants to get this limit should prefer the type-specific methods
     * that return primitives for better performance. This method is only
     * provided for code that has to deal with values in a generic way (without
     * knowing the exact type of the actual value).
     * </p>
     * 
     * @return lower control limit (never <code>null</code>).
     */
    ElementType getGenericLowerControlLimit();

    /**
     * <p>
     * Returns a read-only version of this value. If the value is read-only,
     * this value is returned. Otherwise, a wrapped version of this value, that
     * does not allow modifications is returned.
     * </p>
     * 
     * <p>
     * Please note that the read-only value is not a copy of this value, but
     * just a wrapper around it. This means, that modifications of the wrapped
     * value will result in modifications of the read-only value. Therefore it
     * is bad practice to modify a value that has been used as the base for a
     * read-only value, because code that still has a reference to the read-only
     * value might not expect this change and thus fail unexpectedly.
     * </p>
     * 
     * @return read-only version of this value (never <code>null</code>).
     */
    ChannelAccessNumericControlsValue<ElementType> asReadOnlyValue();

    /**
     * <p>
     * Creates and returns a copy of this object. The object returned is
     * completely independent from this object. This means that modifications to
     * this object will not affect the returned object and vice-versa.
     * </p>
     * 
     * <p>
     * The object returned is guaranteed to be completely identical to this
     * object, including its type. This means that
     * <code>x.clone().equals(x)</code> and
     * <code>x.clone().getClass() == x.getClass()</code> are always
     * <code>true</code>.
     * </p>
     * 
     * <p>
     * The returned value is guaranteed to allow write access, even if this
     * value is read-only. This means that cloning a read-only value is a good
     * way to get a writable copy.
     * </p>
     * 
     * @return a clone of this instance.
     */
    ChannelAccessNumericControlsValue<ElementType> clone();

}
