/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import java.util.ListIterator;

import javax.naming.OperationNotSupportedException;

/**
 * <p>
 * Channel Access value. This interface is the base interface for all values
 * that can be transferred via the Channel Access protocol.
 * </p>
 * 
 * <p>
 * Objects implementing this interface can be expected to have implementations
 * of {@link Object#equals(Object) equals(Object)} and {@link Object#hashCode()
 * hashCode()} that do not simply check for object identity but actually compare
 * the object's internal state. All objects of this type can be cloned.
 * </p>
 * 
 * <p>
 * Objects implementing this interface can be expected to be safe for concurrent
 * access by multiple threads as long as none of these threads modifies the
 * object. Modifications to the object have to be synchronized by establishing
 * an external mutex.
 * </p>
 * 
 * <p>
 * A read-only version of a value can be retrieved through the
 * {@link #asReadOnlyValue()}. Such a version shares it data with the read-write
 * variant that it has been created from as is suitable to be passed to other
 * components that shall not modify the value. Modifications to the original
 * value however will also affect the read-only variants, so typically, code
 * that creates a read-only variant should not keep references to the read-write
 * variant or ensure that it does not modify the read-write variant while other
 * threads are accessing the read-only variant.
 * </p>
 * 
 * <p>
 * All objects that implement this interface can be cloned. Cloning an instance
 * always yields a read-write value, even if cloning a read-only value. The
 * cloned instance is completely independent of its origin, so it can safely be
 * modified without affecting the value it was cloned from.
 * </p>
 * 
 * <p>
 * Channel Access values can be created using {@link ChannelAccessValueFactory}
 * or can be deserialized from their serialized representation using
 * {@link ChannelAccessValueCodec}.
 * </p>
 *
 * <p>
 * This interface should not be implemented by user-defined classes. Code that
 * wants to create an instance implementing this interface should use the
 * methods provided by {@link ChannelAccessValueFactory} instead.
 * </p>
 *
 * @param <ElementType>
 *            type of the actual value elements. For primitives, this is the
 *            boxed variant.
 * 
 * @author Sebastian Marsching
 */
public interface ChannelAccessValue<ElementType> extends Cloneable,
        Iterable<ElementType> {

    /**
     * Returns the identifier for this value's underlying type.
     * 
     * @return type identifier (never <code>null</code>).
     */
    ChannelAccessValueType getType();

    /**
     * Returns the number of elements this value has. In combination with
     * {@link #getGenericValueElement(int)}, this method can be used to iterate
     * over the elements, as an alternative to the {@link #iterator()} method.
     * 
     * @return number of elements this value has.
     * @see #getGenericValueElement(int)
     */
    int getValueSize();

    /**
     * <p>
     * Returns the value element with the specified index.
     * </p>
     * 
     * <p>
     * Code that wants to get elements should prefer to use the
     * <code>getValue()</code> method in the respective child class and access
     * the array through the methods provided by the buffer object for better
     * performance (avoiding boxing and unboxing of primitives). This method is
     * only provided for code that has to deal with values in a generic way
     * (without knowing the exact type of the actual value).
     * </p>
     * 
     * @param index
     *            zero-based index into the element array.
     * @return element of the array at the specified index (never
     *         <code>null</code>).
     * @throws IndexOutOfBoundsException
     *             if the index is equal to or greater than the
     *             {@link #getValueSize() size of the element array}. .
     */
    ElementType getGenericValueElement(int index);

    /**
     * Returns a (read-only) iterator that iterates over the actual value
     * elements.
     * 
     * @return iterator over the value elements.
     */
    ListIterator<ElementType> iterator();

    /**
     * <p>
     * Returns a read-only version of this value. If the value is read-only,
     * this value is returned. Otherwise, a wrapped version of this value, that
     * does not allow modifications is returned.
     * </p>
     * 
     * <p>
     * Please note that the read-only value is not a copy of this value, but
     * just a wrapper around it. This means, that modifications of the wrapped
     * value will result in modifications of the read-only value. Therefore it
     * is bad practice to modify a value that has been used as the base for a
     * read-only value, because code that still has a reference to the read-only
     * value might not expect this change and thus fail unexpectedly.
     * </p>
     * 
     * @return read-only version of this value (never <code>null</code>).
     */
    ChannelAccessValue<ElementType> asReadOnlyValue();

    /**
     * Returns whether this value is read-only. A read-only value may not be
     * modified. Calling a method that would modify the value will result in an
     * {@link OperationNotSupportedException}.
     * 
     * @return <code>true</code> if this value is read-only, <code>false</code>
     *         if it may be modified.
     */
    boolean isReadOnly();

    /**
     * <p>
     * Returns a hash code value for this Channel Access value. This method is
     * supported for the benefit of hash tables. Typically, the hash code is
     * calculated as a combination of the hash codes of the elements and
     * meta-data included in this value object. However, the concrete algorithm
     * is up to the implementing class.
     * </p>
     * 
     * <p>
     * If two Channel Access values are equal (as indicated by the return value
     * of {@link #equals(Object)}), they also have the same hash code. However,
     * the fact that two values have the same hash code does not imply that they
     * are equal. This contract must be honored by all implementations of this
     * method.
     * </p>
     * 
     * @return hash code suitable for use by a hash table.
     */
    @Override
    int hashCode();

    /**
     * <p>
     * Tells whether the specified object represents the same Channel Access
     * value.
     * </p>
     * 
     * <p>
     * Two values are considered equal if they are of the same type and have the
     * same elements and meta-data (if the type contains meta-data).
     * </p>
     * 
     * @param obj
     *            reference object with which this object is compared.
     * @return <code>true</code> if <code>obj</code> is a Channel Access Value
     *         that is of the same type and has the elements and meta-data as
     *         this value, <code>false</code> otherwise.
     */
    @Override
    boolean equals(Object obj);

    /**
     * <p>
     * Creates and returns a copy of this object. The object returned is
     * completely independent from this object. This means that modifications to
     * this object will not affect the returned object and vice-versa.
     * </p>
     * 
     * <p>
     * The object returned is guaranteed to be completely identical to this
     * object, including its type. This means that
     * <code>x.clone().equals(x)</code> and
     * <code>x.clone().getClass() == x.getClass()</code> are always
     * <code>true</code>.
     * </p>
     * 
     * <p>
     * The returned value is guaranteed to allow write access, even if this
     * value is read-only. This means that cloning a read-only value is a good
     * way to get a writable copy.
     * </p>
     * 
     * @return a clone of this instance.
     */
    ChannelAccessValue<ElementType> clone();

    /**
     * Returns a string representation of this Channel Access value. The string
     * representation includes the value type, the value elements, and the
     * meta-data (if values of the specific type contain meta-data at all).
     * 
     * @return string representation of this value.
     */
    @Override
    String toString();

}
